#include <cmath>
#include <cstdlib>

#include "../engine/TextRenderer.h"
#include "../entity/GameWorld.h"
#include "../control/TankControl.h"
#include "../decor/SmokeTrail.h"
#include "../decor/TextFloat.h"
#include "../misc/Random.h"
#include "../misc/TextHandler.h"
#include "../misc/Settings.h"
#include "../objects/Explosion.h"
#include "../objects/Shield.h"
#include "../objects/Bullet.h"
#include "../objects/GroundDebris.h"
#include "../sound/SoundSystem.h"

#include "Tank.h"

/***************************************************************/

GLBitmap	*Tank::bodyBMP = NULL;
GLBitmap	*Tank::gunBMP = NULL;
GLBitmap	*Tank::crosshairBMP = NULL;
int		Tank::count = 0;

/***************************************************************/

Tank::Tank(GLColor clr, double _x, double _y)
	: Entity(_x, _y), color(clr)
{
	shieldColor = GLColor(clr.r / 2 + 0.5, clr.g / 2 + 0.5, clr.b / 2 + 0.5);
	smokeColor = GLColor(1.0, 1.0, 1.0);
	
	gunAngle = randomFloat(20.0, 160.0); gunPower = 0.0;
	gunDirection = 1;
	state = TS_IDLE;
	health = 100.0;
	cooldownCounter = 0.0;
	controller = NULL;
	invincible = false;
	
	if (count == 0)
	{
		bodyBMP = GLBitmap::load("gfx/tanks/body.tga", GLBitmap::TF_MAGIC_PINK);
		gunBMP = GLBitmap::load("gfx/tanks/gun.tga", GLBitmap::TF_NO_ALPHA);
		crosshairBMP = GLBitmap::load("gfx/tanks/crosshair.tga", GLBitmap::TF_MAGIC_PINK);
		
		bodyBMP->setOrigin(0.5, 0.5);
		
		gunBMP->setOrigin(0.5, 1.0);
		gunBMP->setTransparencyMode(GLBitmap::TM_ALPHA_BLEND);
		
		crosshairBMP->setTransparencyMode(GLBitmap::TM_ALPHA_CUT);
		crosshairBMP->setOrigin(0.5, 0.5);
	}
	
	count++;
}


Tank::~Tank()
{
	count--;

	if (count == 0)
	{
		delete bodyBMP; delete gunBMP; delete crosshairBMP;
	}
}

/***************************************************************/

bool Tank::contains(double cx, double cy)
{
	if ( (cy >= y + 0.5) && ( cy <= y + 4.5) && (cx >= x - 12) && (cx <= x + 12) )
		return true;
	
	if ( (cy >= y - 3.5) && ( cy <= y + 0.5) && (cx >= x - 5) && (cx <= x + 5) )
		return true;
	
	return false;
}

/***************************************************************/

void Tank::step(double dt)
{
	if (controller)
	{
		TankBehavior whatToDo = controller->getBehavior(dt);
		switch (whatToDo)
		{
			case TB_SHIELD:
				if (state != TS_COOLDOWN)
				{
					if (Settings::setting[GPL_SHIELD_TIME] > 0)
					{
						gunPower = 0.0;
						parent->add(new Shield(x, y, shieldColor, Settings::setting[GPL_SHIELD_TIME] * 20.0), true);
						
						state = TS_COOLDOWN;
						cooldownCounter = 60.0 * Settings::setting[GPL_DOWN_TIME];
						
						this->doDamage(2.0);
					}
					else
					{
						gunPower = 0.0;
						state = TS_COOLDOWN;
						cooldownCounter = dt;
					}
				}
				break;
				
			case TB_SLOW_SHOOTING:
				if (state == TS_IDLE)
					state = TS_SLOW_SHOOTING;
				break;
				
			case TB_START_SHOOTING:
				if (state == TS_IDLE || state == TS_SLOW_SHOOTING)
					state = TS_SHOOTING;
				break;
				
			case TB_SHOOT:
				if (state == TS_SHOOTING)
				{
					if (parent)
					{
						double vx = std::cos(gunAngle * M_PI / 180) * gunPower * 2;
						double vy = -std::sin(gunAngle * M_PI / 180) * gunPower * 2;
						double shotX = x + vx * 10 / gunPower;
						double shotY = y + vy * 10 / gunPower;
						Bullet *shot = new Bullet(shotX, shotY, vx, vy);
						parent->add(shot, true);
						
						if (Settings::setting[GFX_SMOKE] > 0)
						{
							parent->add(new SmokeTrail(shot, 6.0 / Settings::setting[GFX_SMOKE], 60.0, smokeColor), true);
						}
					}
					
					if (gunAngle > 90.0)
						gunDirection = 1;
					else
						gunDirection = -1;
					
					state = TS_COOLDOWN;
					cooldownCounter = 100.0 * Settings::setting[GPL_DOWN_TIME];
					
					gunPower = 0.0;
					
					soundSys->playSound(SND_SHOOT, (int)((x - 400.0) * 256.0 / 400.0) + 128, 0.2);
				}
				break;
			default:
				break;
		}
	}
	
	if (state == TS_IDLE)
	{
		gunAngle += gunDirection * dt * 0.5;
		if (gunAngle > 160.0)
		{
			gunDirection = -1;
			gunAngle = 160.0 - (gunAngle - 160.0);
		}
		if (gunAngle < 20.0)
		{
			gunDirection = 1;
			gunAngle = 20.0 + (20.0 - gunAngle);
		}
	}
	
	if (state == TS_SHOOTING)
	{
		if (gunPower < 1.5)
			gunPower += 0.005 * dt;
	};
	
	if (state == TS_SLOW_SHOOTING)
	{
		if (gunPower < 1.5)
			gunPower += 0.001 * dt;
	}
	
	if (state == TS_COOLDOWN)
	{
		gunAngle += gunDirection * dt * 0.2;

		if (gunAngle > 160.0)
			gunAngle = 160.0;
		if (gunAngle < 20.0)
			gunAngle = 20.0;

		cooldownCounter -= dt;
		if (cooldownCounter <= 0.0)
			state = TS_IDLE;
	}
}


void Tank::render()
{
	gunBMP->setTint(color);
	gunBMP->setRotation(gunAngle - 90.0);
	gunBMP->draw(x, y);
	
	bodyBMP->setTint(color);
	bodyBMP->draw(x, y);
	
	if (state != TS_COOLDOWN)
	{
		double chX, chY;
		chX = (int)(x + std::cos(gunAngle * M_PI / 180) * (60.0 - gunPower * 30.0)) + 0.5;
		chY = (int)(y - std::sin(gunAngle * M_PI / 180) * (60.0 - gunPower * 30.0)) + 0.5;
		crosshairBMP->setTint(color);
		crosshairBMP->draw(chX, chY);
	}
	
	if (health > 0.0)
	{
		GLColor healthColor;
		if (health > 50.0)
		{
			healthColor.r = (100.0 - health) / 50.0;
			healthColor.g = ((health - 50.0) * 0.6 + (100 - health) * 1.0) / 50.0;
		}
		else
		{
			healthColor.r = 1.0;
			healthColor.g = health / 50.0;
		}
		healthColor.b = 0.0;
		
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_QUADS);
			glColor3f(0.0, 0.0, 0.0);
			glVertex2f(x - 10, y + 8); glVertex2f(x + 10, y + 8);
			glVertex2f(x + 10, y + 12); glVertex2f(x - 10, y + 12);
			glColor3f(healthColor.r, healthColor.g, healthColor.b);
			glVertex2f(x - 9, y + 9); glVertex2f(x - 9 + 18 * health / 100, y + 9);
			glVertex2f(x - 9 + 18 * health / 100, y + 11); glVertex2f(x - 9, y + 11);
		glEnd();
	}
}

/********************************************/

void Tank::doDamage(double dmg)
{
	if (health <= 0.0) return;
	if (invincible) return;
		
	health -= dmg;
	
	char buf[8];
	sprintf(buf, "%d", (int)dmg);
	
	parent->add(new TextFloat(x, y, buf, GLColor(1.0, 0.5, 0.5), 0.32), true);
	
	if (health <= 0.0)
	{
		parent->add(new Explosion(x, y, 60.0, 0.0, 0.0, 400.0), true);
		for (int i = 0; i < Settings::setting[GFX_DEBRIS] * 15; i++)
			parent->add(new GroundDebris(x, y, 0.5), true);
		
		soundSys->playSound(SND_BOOM, (int)((x - 400.0) * 256.0 / 400.0) + 128, 0.2);
	}
}
